@@ -3,10 +3,14 @@ import pwseqdist as pw
 import numpy as np
 import warnings
 from . import repertoire_db
-from tcrdist.rep_funcs import _pws
+from tcrdist.rep_funcs import _pws, compute_pws_sparse
+from tcrdist.sample import _default_sampler
+from tcrdist.background import get_stratified_gene_usage_frequency
+from tcrdist.background import _synthesize_human_beta_vj_background, _synthesize_human_alpha_vj_background, _synthesize_mouse_beta_vj_background
 from zipdist.zip2 import Zipdist2
 import sys
 
+
 class TCRrep:
     """
     Flexible distance measures for comparing T cell receptors
@@ -149,8 +153,9 @@ class TCRrep:
             self.clone_df = clone_df 
 
             if cell_df is None:
-                cell_df = pd.DataFrame()
-            self.cell_df = cell_df
+                cell_df = pd.DataFrame()   
+            self.cell_df = cell_df            
+           
             self._validate_cell_df()
             self.clone_df = clone_df
             self.df2 = df2
@@ -191,6 +196,9 @@ class TCRrep:
                 self._initialize_chain_specific_attributes()
         
             if compute_distances:
+                # This is a safety, measure so that a new user doesn't accidently try to compute a pairwise matrix that won't fit in memory
+                if self.clone_df.shape[0] > 10000:
+                    warnings.warn(f"<clone_df> size {self.clone_df.shape[0]} > 10,000. TCRrep.compute_distances() must be called explicitly by user with knowledge of system memory availability")
                 self.compute_distances()
 
             if self.archive_result:
@@ -330,51 +338,57 @@ class TCRrep:
             self._assign_distance_attributes(d = pw_delta, chain = 'delta')
     
     def compute_rect_distances(self, df = None, df2 = None, store = None):
+        self._rect_distances(pw_dist_func=_pws, df = df, df2 = df2, store = store)
+    
+    def compute_sparse_rect_distances(self, df=None, df2=None, radius=50, chunk_size=500):
+        self._rect_distances(pw_dist_func=compute_pws_sparse, df=df, df2=df2, radius=radius, chunk_size=chunk_size)
+    
+    def _rect_distances(self, pw_dist_func, df = None, df2 = None, **kwargs):
         if df is None:
             df = self.clone_df
         if df2 is None:
             df2 = self.df2
-        if store is None:
-            store = self.store_all_cdr
+        if not 'store' in kwargs or kwargs['store'] is None:
+            kwargs['store'] = self.store_all_cdr
         if 'alpha' in self.chains:
-            pw_alpha  = _pws(
+            pw_alpha  = pw_dist_func(
                 df = df,
                 df2 = df2,
                 metrics = self.metrics_a, 
                 weights = self.weights_a, 
                 kargs   = self.kargs_a, 
                 cpu     = self.cpus, 
-                store   = store)
+                **kwargs)
             self._assign_distance_attributes(d = pw_alpha, chain = 'alpha',prefix = "rw")
         if 'beta' in self.chains:
-            pw_beta = _pws(
+            pw_beta = pw_dist_func(
                 df = df,
                 df2 = df2, 
                 metrics = self.metrics_b, 
                 weights = self.weights_b, 
                 kargs   = self.kargs_b, 
                 cpu     = self.cpus, 
-                store   = store)
+                **kwargs)
             self._assign_distance_attributes(d = pw_beta, chain = 'beta',prefix = "rw")
         if 'gamma' in self.chains:
-            pw_gamma = _pws(
+            pw_gamma = pw_dist_func(
                 df = df,
                 df2 = df2, 
                 metrics = self.metrics_g, 
                 weights = self.weights_g, 
                 kargs   = self.kargs_g, 
                 cpu     = self.cpus, 
-                store   = store)
+                **kwargs)
             self._assign_distance_attributes(d = pw_gamma, chain = 'gamma', prefix = "rw")
         if 'delta' in self.chains:
-            pw_delta = _pws(
+            pw_delta = pw_dist_func(
                 df = df,
                 df2 = df2,
                 metrics = self.metrics_d,
                 weights = self.weights_d, 
                 kargs   = self.kargs_d, 
                 cpu     = self.cpus, 
-                store   = store)
+                **kwargs)
             self._assign_distance_attributes(d = pw_delta, chain = 'delta', prefix = "rw")
 
     def _assign_distance_attributes(self, d:dict, chain:str, prefix = 'pw'):
@@ -462,21 +476,21 @@ class TCRrep:
                                                              organism = self.organism,
                                                              attr ='cdrs')
         if chain == "alpha":
-            self.cell_df['cdr1_a_aa'] = list(map(f0, self.cell_df.v_a_gene))
-            self.cell_df['cdr2_a_aa'] = list(map(f1, self.cell_df.v_a_gene))
-            self.cell_df['pmhc_a_aa'] = list(map(f2, self.cell_df.v_a_gene))
+            self.cell_df = self.cell_df.assign(cdr1_a_aa=list(map(f0, self.cell_df.v_a_gene)),
+                                               cdr2_a_aa=list(map(f1, self.cell_df.v_a_gene)),
+                                               pmhc_a_aa=list(map(f2, self.cell_df.v_a_gene)))
         if chain == "beta":
-            self.cell_df['cdr1_b_aa'] = list(map(f0, self.cell_df.v_b_gene))
-            self.cell_df['cdr2_b_aa'] = list(map(f1, self.cell_df.v_b_gene))
-            self.cell_df['pmhc_b_aa'] = list(map(f2, self.cell_df.v_b_gene))
+            self.cell_df = self.cell_df.assign(cdr1_b_aa=list(map(f0, self.cell_df.v_b_gene)),
+                                               cdr2_b_aa=list(map(f1, self.cell_df.v_b_gene)),
+                                               pmhc_b_aa=list(map(f2, self.cell_df.v_b_gene)))
         if chain == "gamma":
-            self.cell_df['cdr1_g_aa'] = list(map(f0, self.cell_df.v_g_gene))
-            self.cell_df['cdr2_g_aa'] = list(map(f1, self.cell_df.v_g_gene))
-            self.cell_df['pmhc_g_aa'] = list(map(f2, self.cell_df.v_g_gene))
+            self.cell_df = self.cell_df.assign(cdr1_g_aa=list(map(f0, self.cell_df.v_g_gene)),
+                                               cdr2_g_aa=list(map(f1, self.cell_df.v_g_gene)),
+                                               pmhc_g_aa=list(map(f2, self.cell_df.v_g_gene)))
         if chain == "delta":
-            self.cell_df['cdr1_d_aa'] = list(map(f0, self.cell_df.v_d_gene))
-            self.cell_df['cdr2_d_aa'] = list(map(f1, self.cell_df.v_d_gene))
-            self.cell_df['pmhc_d_aa'] = list(map(f2, self.cell_df.v_d_gene))
+            self.cell_df = self.cell_df.assign(cdr1_d_aa=list(map(f0, self.cell_df.v_d_gene)),
+                                               cdr2_d_aa=list(map(f1, self.cell_df.v_d_gene)),
+                                               pmhc_d_aa=list(map(f2, self.cell_df.v_d_gene)))
     
     def infer_index_cols(self):
         """
@@ -494,7 +508,10 @@ class TCRrep:
         if np.sum(self.cell_df['count']) != np.sum(clone_df['count']):
             n_cells_lost = np.sum(self.cell_df['count']) - np.sum(clone_df['count'])
             n_cell = np.sum(self.cell_df['count'])
-            warnings.warn(f"Not all cells/sequences could be grouped into clones. {n_cells_lost} of {n_cell} were not captured. This occurs when any of the values in the index columns are null or missing for a given sequence. To see entries with missing values use: tcrdist.repertoire.TCRrep._show_incomplete()\n")
+            warnings.warn(f"Not all cells/sequences could be grouped into clones."
+                          f"{n_cells_lost} of {n_cell} were not captured. This occurs when "
+                          "any of the values in the index columns are null or missing for a given sequence. "
+                          "To see entries with missing values use: tcrdist.repertoire.TCRrep._show_incomplete()", stacklevel=2)
         
         # if no clone id column provided thetrn create one as a sequence of numbers
         if "clone_id" not in clone_df:
@@ -629,6 +646,40 @@ class TCRrep:
                 "cdr1_d_aa" : {'use_numba': True, 'distance_matrix': pw.matrices.tcr_nb_distance_matrix, 'dist_weight': 1, 'gap_penalty':4, 'ntrim':0, 'ctrim':0, 'fixed_gappos':True}}
       
 
+    def synthesize_vj_matched_background(self, ts  = None, chain = "beta"):
+        """
+        tcrsampler : TCRsampler or None
+       
+        chain : str 
+            'beta' (in future, TODO: add 'alpha')
+        TODO
+        -------
+        ONLY WORKS CURRENTLY FOR HUMAN BETA, VIA OLGA
+        """
+        if chain not in ["beta", "alpha"]:
+            raise ValueError("Invalid <chain> argument.")
+        if chain == "beta":
+            if ts is None:
+                ts = _default_sampler(organism = self.organism, chain = "beta")()
+                ts = get_stratified_gene_usage_frequency(ts = ts, replace = True) 
+            if self.organism == "human":
+                vj_background = _synthesize_human_beta_vj_background(ts = ts, df = self.clone_df)
+            elif self.organism == "mouse":
+                vj_background = _synthesize_mouse_beta_vj_background(ts = ts, df = self.clone_df)
+        # TODO: ADD OTHER OPTIONS
+        elif chain == "alpha":
+            if self.organism == "human":
+                raise ValueError("TODO: FUTURE VERSIONS NEED ALPHA(HUMAN)")
+                #vj_background = _synthesize_human_alpha_vj_background(ts = ts, df = self.clone_df)
+            elif self.organism == "mouse":
+                raise ValueError("TODO: FUTURE VERSIONS NEED ALPHA(MOUSE)")
+                #vj_background = _synthesize_alpha_beta_vj_background(ts = ts, df = self.clone_df)
+        return vj_background   
+
+
+
+
+
     def _map_gene_to_reference_seq2(self,
                                     organism,
                                     gene,
@@ -653,7 +704,7 @@ class TCRrep:
             aa_string = self.all_genes[organism][gene].__dict__[attr][cdr]
         except KeyError:
             aa_string = None
-            warnings.warn("{} gene was not recognized in reference db no cdr seq could be inferred".format(gene))
+            warnings.warn("{} gene was not recognized in reference db no cdr seq could be inferred".format(gene), stacklevel=2)
         return(aa_string)
        
         """
@@ -665,7 +716,7 @@ class TCRrep:
         Issues warning if invalid organism is passed to TCRrep __init__
         """
         if self.db_file not in ['alphabeta_gammadelta_db.tsv','alphabeta_db.tsv','gammadelta_db.tsv']:
-            warnings.warn("db_file must be 'alphabeta_gammadelta_db.tsv' or 'alphabeta_db.tsv' or 'gammadelta_db.tsv' unless you have built tcrdist3 from scratch")
+            warnings.warn("db_file must be 'alphabeta_gammadelta_db.tsv' or 'alphabeta_db.tsv' or 'gammadelta_db.tsv' unless you have built tcrdist3 from scratch", stacklevel=2)
     
     def _validate_organism(self):
         """
@@ -691,34 +742,35 @@ class TCRrep:
         Issues warning if TCRrep.cell_df not properly formatted.
         """
         if not isinstance(self.cell_df, pd.DataFrame):
-            warnings.warn('TCRrep cell_df argument must be pandas.DataFrame unless you are providing a clone_df directly\n')
+            warnings.warn('TCRrep cell_df argument must be pandas.DataFrame unless you are providing a clone_df directly\n', stacklevel=2)
 
         else:
             cell_df_columns = self.cell_df.columns.to_list()
             
             if "count" not in cell_df_columns:
-                warnings.warn("cell_df needs a counts column to track clonal number of frequency\n")
-            
+                warnings.warn("cell_df needs a counts column to track clonal number of frequency\n", stacklevel=2)
+                warnings.warn("No 'count' column provided; count column set to 1")
+                self.cell_df['count'] = 1
             if "alpha" in self.chains:
                 if not "cdr3_a_aa" in cell_df_columns:
-                    warnings.warn("cell_df needs a column called 'cdr3_a_aa' to track the CDR3 amino acid sequence\n")
+                    warnings.warn("cell_df needs a column called 'cdr3_a_aa' to track the CDR3 amino acid sequence\n", stacklevel=2)
                 if not "v_a_gene" in cell_df_columns:
-                    warnings.warn("cell_df needs a column called 'v_a_gene' for default functions\n")
+                    warnings.warn("cell_df needs a column called 'v_a_gene' for default functions\n", stacklevel=2)
             if "beta" in self.chains:
                 if not "cdr3_b_aa" in cell_df_columns:
-                    warnings.warn("cell_df needs a column called 'cdr3_b_aa' to track the CDR3 amino acid sequence\n")
+                    warnings.warn("cell_df needs a column called 'cdr3_b_aa' to track the CDR3 amino acid sequence\n", stacklevel=2)
                 if not "v_b_gene" in cell_df_columns:
-                    warnings.warn("cell_df needs a column called 'v_b_gene' for default functions\n")
+                    warnings.warn("cell_df needs a column called 'v_b_gene' for default functions\n", stacklevel=2)
             if "gamma" in self.chains:
                 if not "cdr3_g_aa" in cell_df_columns:
-                    warnings.warn("cell_df needs a column called 'cdr3_g_aa' to track the CDR3 amino acid sequence\n")
+                    warnings.warn("cell_df needs a column called 'cdr3_g_aa' to track the CDR3 amino acid sequence\n", stacklevel=2)
                 if not "v_g_gene" in cell_df_columns:
-                    warnings.warn("cell_df needs a column called 'v_g_gene' for default functions\n")
+                    warnings.warn("cell_df needs a column called 'v_g_gene' for default functions\n", stacklevel=2)
             if "delta" in self.chains:
                 if not "cdr3_d_aa" in cell_df_columns:
-                    warnings.warn("cell_df needs a column called 'cdr3_d_aa' to track the CDR3 amino acid sequence\n")
+                    warnings.warn("cell_df needs a column called 'cdr3_d_aa' to track the CDR3 amino acid sequence\n", stacklevel=2)
                 if not "v_d_gene" in cell_df_columns:
-                    warnings.warn("cell_df needs a column called 'v_d_gene' for default functions\n")
+                    warnings.warn("cell_df needs a column called 'v_d_gene' for default functions\n", stacklevel=2)
 
     def _validate_imgt_aligned(self):
         """
